@@ -17,8 +17,10 @@ import io
 import os
 import os.path as osp
 import pdb
+import re
 import shlex
 import sys
+import sysconfig
 import time
 import warnings
 
@@ -127,36 +129,6 @@ if os.name == 'nt' and PY2:
         pass
 
 
-#==============================================================================
-# Cython support
-#==============================================================================
-RUN_CYTHON = os.environ.get("SPY_RUN_CYTHON") == "True"
-HAS_CYTHON = False
-
-if RUN_CYTHON:
-    try:
-        __import__('Cython')
-        HAS_CYTHON = True
-    except Exception:
-        pass
-
-    if HAS_CYTHON:
-        # Import pyximport to enable Cython files support for
-        # import statement
-        import pyximport
-        pyx_setup_args = {}
-
-        # Add Numpy include dir to pyximport/distutils
-        try:
-            import numpy
-            pyx_setup_args['include_dirs'] = numpy.get_include()
-        except Exception:
-            pass
-
-        # Setup pyximport and enable Cython files reload
-        pyximport.install(setup_args=pyx_setup_args, reload_support=True)
-
-
 #==============================================================================
 # Prevent subprocess.Popen calls to create visible console windows on Windows.
 # See issue #4932
@@ -230,6 +202,33 @@ class IPyTesProgram(TestProgram):
         TestProgram.__init__(self, *args, **kwargs)
 unittest.main = IPyTesProgram
 
+# Ignore some IPython/ipykernel warnings
+try:
+    warnings.filterwarnings(action='ignore', category=DeprecationWarning,
+                            module='ipykernel.ipkernel')
+except:
+    pass
+
+#==============================================================================
+# Turtle adjustments
+#==============================================================================
+# This is needed to prevent turtle scripts crashes after multiple runs in the
+# same IPython Console instance.
+# See Spyder issue #6278
+try:
+    import turtle
+    from turtle import Screen, Terminator
+
+    def spyder_bye():
+        try:
+            Screen().bye()
+            turtle.TurtleScreen._RUNNING = True
+        except Terminator:
+            pass
+    turtle.bye = spyder_bye
+except:
+    pass
+
 
 #==============================================================================
 # Pandas adjustments
@@ -272,6 +271,35 @@ except:
     pass
 
 
+# =============================================================================
+# Multiprocessing adjustments
+# =============================================================================
+# This patch is only needed on Windows and Python 3
+if os.name == 'nt' and not PY2:
+    # This could fail with changes in Python itself, so we protect it
+    # with a try/except
+    try:
+        import multiprocessing.spawn
+        _old_preparation_data = multiprocessing.spawn.get_preparation_data
+
+        def _patched_preparation_data(name):
+            """
+            Patched get_preparation_data to work when all variables are
+            removed before execution.
+            """
+            try:
+                return _old_preparation_data(name)
+            except AttributeError:
+                main_module = sys.modules['__main__']
+                # Any string for __spec__ does the job
+                main_module.__spec__ = ''
+                return _old_preparation_data(name)
+
+        multiprocessing.spawn.get_preparation_data = _patched_preparation_data
+    except Exception:
+        pass
+
+
 #==============================================================================
 # Pdb adjustments
 #==============================================================================
@@ -496,14 +524,25 @@ class UserModuleReloader(object):
     def __init__(self, namelist=None, pathlist=None):
         if namelist is None:
             namelist = []
-        spy_modules = ['sitecustomize', 'spyder', 'spyderplugins']
+        else:
+            try:
+                namelist = namelist.split(',')
+            except Exception:
+                namelist = []
+
+        # Spyder modules
+        spy_modules = ['spyder_kernels']
+
+        # Matplotlib modules
         mpl_modules = ['matplotlib', 'tkinter', 'Tkinter']
+
         # Add other, necessary modules to the UMR blacklist
         # astropy: see issue 6962
         # pytorch: see issue 7041
         # fastmat: see issue 7190
         # pythoncom: see issue 7190
-        other_modules = ['pytorch', 'pythoncom']
+        # tensorflow: see issue 8697
+        other_modules = ['pytorch', 'pythoncom', 'tensorflow']
         if PY2:
             py2_modules = ['astropy', 'fastmat']
             other_modules = other_modules + py2_modules
@@ -511,44 +550,171 @@ class UserModuleReloader(object):
 
         if pathlist is None:
             pathlist = []
-        self.pathlist = pathlist
+        self.pathlist = self.create_pathlist(pathlist)
+
+        # List of previously loaded modules
         self.previous_modules = list(sys.modules.keys())
 
-    def is_module_blacklisted(self, modname, modpath):
-        if HAS_CYTHON:
-            # Don't return cached inline compiled .PYX files
-            return True
-        for path in [sys.prefix]+self.pathlist:
-            if modpath.startswith(path):
-                return True
-        else:
-            return set(modname.split('.')) & set(self.namelist)
+        # List of module names to reload
+        self.modnames_to_reload = []
 
-    def run(self, verbose=False):
+        # Activate Cython support
+        self.has_cython = False
+        self.activate_cython()
+
+        # Check if the UMR is enabled or not
+        enabled = os.environ.get("SPY_UMR_ENABLED", "")
+        self.enabled = enabled.lower() == "true"
+
+        # Check if the UMR should print the list of reloaded modules or not
+        verbose = os.environ.get("SPY_UMR_VERBOSE", "")
+        self.verbose = verbose.lower() == "true"
+
+    def create_pathlist(self, initial_pathlist):
         """
-        Del user modules to force Python to deeply reload them
+        Add to pathlist Python library paths to be skipped from module
+        reloading.
+        """
+        # Get standard installation paths
+        try:
+            paths = sysconfig.get_paths()
+            standard_paths = [paths['stdlib'],
+                              paths['purelib'],
+                              paths['scripts'],
+                              paths['data']]
+        except Exception:
+            standard_paths = []
+
+        # Get user installation path
+        # See Spyder issue 8776
+        try:
+            import site
+            if getattr(site, 'getusersitepackages', False):
+                # Virtualenvs don't have this function but
+                # conda envs do
+                user_path = [site.getusersitepackages()]
+            elif getattr(site, 'USER_SITE', False):
+                # However, it seems virtualenvs have this
+                # constant
+                user_path = [site.USER_SITE]
+            else:
+                user_path = []
+        except Exception:
+            user_path = []
+
+        return initial_pathlist + standard_paths + user_path
+
+    def is_module_reloadable(self, module, modname):
+        """Decide if a module is reloadable or not."""
+        if self.has_cython:
+            # Don't return cached inline compiled .PYX files
+            return False
+        else:
+            if (self.is_module_in_pathlist(module) or
+                    self.is_module_in_namelist(modname)):
+                return False
+            else:
+                return True
+
+    def is_module_in_namelist(self, modname):
+        """Decide if a module can be reloaded or not according to its name."""
+        return set(modname.split('.')) & set(self.namelist)
+
+    def is_module_in_pathlist(self, module):
+        """Decide if a module can be reloaded or not according to its path."""
+        modpath = getattr(module, '__file__', None)
+
+        # Skip module according to different criteria
+        if modpath is None:
+            # *module* is a C module that is statically linked into the
+            # interpreter. There is no way to know its path, so we
+            # choose to ignore it.
+            return True
+        elif any([p in modpath for p in self.pathlist]):
+            # We don't want to reload modules that belong to the
+            # standard library or installed to site-packages,
+            # just modules created by the user.
+            return True
+        elif not os.name == 'nt':
+            # Module paths containing the strings below can be ihherited
+            # from the default Linux installation, Homebrew or the user
+            # site-packages in a virtualenv.
+            patterns = [r'^/usr/lib.*',
+                        r'^/usr/local/lib.*',
+                        r'^/usr/.*/dist-packages/.*',
+                        r'^/home/.*/.local/lib.*',
+                        r'^/Library/.*',
+                        r'^/Users/.*/Library/.*',
+                        r'^/Users/.*/.local/.*',
+            ]
+
+            if [p for p in patterns if re.search(p, modpath)]:
+                return True
+            else:
+                return False
+        else:
+            return False
+
+    def activate_cython(self):
+        """
+        Activate Cython support.
+
+        We need to run this here because if the support is
+        active, we don't to run the UMR at all.
+        """
+        run_cython = os.environ.get("SPY_RUN_CYTHON") == "True"
+
+        if run_cython:
+            try:
+                __import__('Cython')
+                self.has_cython = True
+            except Exception:
+                pass
+
+            if self.has_cython:
+                # Import pyximport to enable Cython files support for
+                # import statement
+                import pyximport
+                pyx_setup_args = {}
+
+                # Add Numpy include dir to pyximport/distutils
+                try:
+                    import numpy
+                    pyx_setup_args['include_dirs'] = numpy.get_include()
+                except Exception:
+                    pass
+
+                # Setup pyximport and enable Cython files reload
+                pyximport.install(setup_args=pyx_setup_args,
+                                  reload_support=True)
+
+    def run(self):
+        """
+        Delete user modules to force Python to deeply reload them
 
         Do not del modules which are considered as system modules, i.e.
         modules installed in subdirectories of Python interpreter's binary
         Do not del C modules
         """
-        log = []
+        self.modnames_to_reload = []
         for modname, module in list(sys.modules.items()):
             if modname not in self.previous_modules:
-                modpath = getattr(module, '__file__', None)
-                if modpath is None:
-                    # *module* is a C module that is statically linked into the
-                    # interpreter. There is no way to know its path, so we
-                    # choose to ignore it.
-                    continue
-                if not self.is_module_blacklisted(modname, modpath):
-                    log.append(modname)
+                # Decide if a module can be reloaded or not
+                if self.is_module_reloadable(module, modname):
+                    self.modnames_to_reload.append(modname)
                     del sys.modules[modname]
-        if verbose and log:
-            _print("\x1b[4;33m%s\x1b[24m%s\x1b[0m"\
-                   % ("Reloaded modules", ": "+", ".join(log)))
+                else:
+                    continue
 
-__umr__ = None
+        # Report reloaded modules
+        if self.verbose and self.modnames_to_reload:
+            modnames = self.modnames_to_reload
+            _print("\x1b[4;33m%s\x1b[24m%s\x1b[0m"\
+                   % ("Reloaded modules", ": "+", ".join(modnames)))
+
+
+__umr__ = UserModuleReloader(namelist=os.environ.get("SPY_UMR_NAMELIST",
+                                                     None))
 
 
 #==============================================================================
@@ -630,16 +796,9 @@ def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
         # UnicodeError, TypeError --> eventually raised in Python 2
         # AttributeError --> systematically raised in Python 3
         pass
-    global __umr__
-    if os.environ.get("SPY_UMR_ENABLED", "").lower() == "true":
-        if __umr__ is None:
-            namelist = os.environ.get("SPY_UMR_NAMELIST", None)
-            if namelist is not None:
-                namelist = namelist.split(',')
-            __umr__ = UserModuleReloader(namelist=namelist)
-        else:
-            verbose = os.environ.get("SPY_UMR_VERBOSE", "").lower() == "true"
-            __umr__.run(verbose=verbose)
+
+    if __umr__.enabled:
+        __umr__.run()
     if args is not None and not isinstance(args, basestring):
         raise TypeError("expected a character buffer object")
     if namespace is None:
@@ -659,7 +818,7 @@ def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
         os.chdir(wdir)
     if post_mortem:
         set_post_mortem()
-    if HAS_CYTHON:
+    if __umr__.has_cython:
         # Cython files
         with io.open(filename, encoding='utf-8') as f:
             ipython_shell = get_ipython()
@@ -669,7 +828,13 @@ def runfile(filename, args=None, wdir=None, namespace=None, post_mortem=False):
 
     clear_post_mortem()
     sys.argv = ['']
-    namespace.pop('__file__')
+
+    # Avoid error when running `%reset -f` programmatically
+    # See issue spyder-ide/spyder-kernels#91
+    try:
+        namespace.pop('__file__')
+    except KeyError:
+        pass
 
 builtins.runfile = runfile
 
